from Products.Five import BrowserView
from collective.mcp import categories, pages
import logging

logger = logging.getLogger('collective.multimodeview')


class MacControlPanel(BrowserView):
    def base_url(self):
        return '%s/%s' % (self.context.absolute_url(),
                          self.__name__)

    def filter_pages(self):
        """ Provides a list of pages that are
        viewable by the user.
        """
        filtered_pages = []
        used_ids = []

        for pclass in pages:
            try:
                page = self.context.restrictedTraverse(
                    '@@%s' % pclass.zcml_id)
            except:
                # The page does not exist in this context or is not
                # accessible for the user.
                logger.info('Page "%s" not found' % pclass.zcml_id)
                continue

            page.view = self
            try:
                page_id = page.widget_id

                if page_id in used_ids:
                    msg = "widget_id '%s' is already used. Choose another one."
                    logger.warn(msg % page_id)
                    continue

                if not page.is_shown():
                    logger.info('Page "%s" not visible' % pclass.zcml_id)
                    continue

                filtered_pages.append(page)
                used_ids.append(page_id)

            except AttributeError:
                logger.warn(
                    "You did not provide 'widget_id' for class %s" %
                    pclass)

        return filtered_pages

    def __call__(self):
        pages = self.filter_pages()

        form = self.request.form
        requested_id = form.get('widget_id', None)

        for page in pages:
            if requested_id == page.widget_id:
                # Ok we'll just render the page requested.
                rendered = page()

                # There is a special mode for sub-page called 'back'.
                # In this mode, the page is not shown but the main menu.
                if page.mode != 'back':
                    self.sub_page = page
                    self.sub_page_rendered = rendered
                    return self.index()

                # Well, no need to look for other pages.
                break

        # No page (or an incorrect) page was requested, so we
        # render the default view with all the icons.
        # We build a temp dictionnary to store
        # categories and page.
        # This will then be treated for:
        # - keeping the same order than the original categories list.
        # - removing the empty categories
        categories_tmp = {}

        for page in pages:
            page_cat = page.category
            if page_cat not in categories_tmp:
                categories_tmp[page_cat] = []

            categories_tmp[page_cat].append(page)

        # Now be build the real 'categories' attribute that will be used
        # to render the main page.
        self.categories = []

        for cat in categories:
            if cat.id not in categories_tmp:
                # Well, there's no page for this category.
                continue

            cat.context = self.context
            self.categories.append({'category': cat,
                                    'pages': categories_tmp[cat.id]
                                    })

        return self.index()



# Customized versions of the pages defined in notes.py and home_message
# with restricted accesses.

from collective.mcp.samples.notes import Notes
from collective.mcp.samples.home_message import HomeMessage


class HomeMessageRestricted(HomeMessage):
    zcml_id = 'collective_mcp_home_message_restricted'
    widget_id = 'collective_mcp_home_message_restricted'
    title = 'Home message (restricted to admin)'


class HomeMessageRestrictedII(HomeMessage):
    zcml_id = 'collective_mcp_home_message_restrictedII'
    widget_id = 'collective_mcp_home_message_restrictedII'
    title = 'Home message (restricted using "permission" attribute)'

    view_permission = 'Manage portal'


class HomeMessageRestrictedIII(HomeMessage):
    zcml_id = 'collective_mcp_home_message_restrictedIII'
    widget_id = 'collective_mcp_home_message_restrictedIII'
    title = 'Home message (restricted using "is_shown" method)'

    def is_shown(self):
        """ Now the page is only visible is you are not a manager.
        Yes it's not a real-life use case.
        """
        user = self.get_user()
        if user:
            return 'Manager' not in user.getRolesInContext(
                self.context)
        return True


class NotesRestricted(Notes):
    zcml_id = 'collective_mcp_notes_restricted'
    widget_id = 'collective_mcp_notes_restricted'
    title = 'Partially restricted notes'

    @property
    def modes(self):
        modes = {'add': {'success_msg': 'The note has been added',
                         'error_msg': 'Impossible to add a note: please correct the form',
                         'submit_label': 'Add note'},
                 'edit': {'success_msg': 'The note has been edited',
                          'submit_label': 'Edit note'}}

        if self.check_permission('Manage portal'):
            modes['delete'] = {'success_msg': 'The note has been deleted',
                               'submit_label': 'Add note'}
        return modes

    @property
    def multi_objects_buttons(self):
        buttons = ['add']
        if self.check_permission('Manage portal'):
            buttons.append('delete')

        return buttons



from Products.Five import fiveconfigure, zcml
from Products.PloneTestCase import PloneTestCase as ptc
from Products.PloneTestCase.layer import PloneSite
from Testing import ZopeTestCase as ztc
from collective.mcp.sorted_list import SortedList
import collective.mcp
import collective.multimodeview
import doctest
import transaction
import unittest

ptc.setupPloneSite()
OPTIONFLAGS = (doctest.ELLIPSIS |
               doctest.NORMALIZE_WHITESPACE)


class TestCase(ptc.PloneTestCase):
    def __init__(self, *args, **kwargs):
        super(TestCase, self).__init__(*args, **kwargs)

        from Products.Five.testbrowser import Browser
        self.browser = Browser()

    def login_as_user(self, username, password):
        self.browser.open('http://nohost/plone/logout')
        self.browser.open('http://nohost/plone/login_form')
        self.browser.getControl(name='__ac_name').value = username
        self.browser.getControl(name='__ac_password').value = password
        self.browser.getControl(name='submit').click()

    def login_as_manager(self):
        self.login_as_user(
            ptc.portal_owner,
            ptc.default_password)

    class layer(PloneSite):

        @classmethod
        def setUp(cls):
            fiveconfigure.debug_mode = True
            zcml.load_config('configure.zcml',
                             collective.multimodeview)
            ztc.installPackage(collective.multimodeview)

            zcml.load_config('configure.zcml',
                             collective.mcp)
            zcml.load_config('samples.zcml',
                             collective.mcp)
            ztc.installPackage(collective.mcp)

            # We need to clean what has been added by samples.
            collective.mcp.categories = SortedList(id_attr='id')
            collective.mcp.pages = SortedList(id_attr='widget_id')

            fiveconfigure.debug_mode = False

        @classmethod
        def tearDown(cls):
            pass

    def afterSetUp(self):
        # his hack allows us to get the traceback when the getting a
        # 500 error when using the browser.
        self.portal.error_log._ignored_exceptions = ()

        def raising(self, info):
            import traceback
            traceback.print_tb(info[2])
            print info[1]

        from Products.SiteErrorLog.SiteErrorLog import SiteErrorLog
        SiteErrorLog.raising = raising
        transaction.commit()


def test_suite():
    return unittest.TestSuite([
        ztc.FunctionalDocFileSuite(
            'doc/main.rst',
            package='collective.mcp',
            optionflags=OPTIONFLAGS,
            test_class=TestCase),
        ztc.FunctionalDocFileSuite(
            'doc/restriction.rst',
            package='collective.mcp',
            optionflags=OPTIONFLAGS,
            test_class=TestCase),
        ztc.FunctionalDocFileSuite(
            'doc/defect.rst',
            package='collective.mcp',
            optionflags=OPTIONFLAGS,
            test_class=TestCase),
        ztc.FunctionalDocFileSuite(
            'doc/modes.rst',
            package='collective.mcp',
            optionflags=OPTIONFLAGS,
            test_class=TestCase),
        ztc.FunctionalDocFileSuite(
            'doc/multiobjects.rst',
            package='collective.mcp',
            optionflags=OPTIONFLAGS,
            test_class=TestCase),
        ztc.FunctionalDocFileSuite(
            'doc/theming.rst',
            package='collective.mcp',
            optionflags=OPTIONFLAGS,
            test_class=TestCase),
    ])



from collective.mcp import Category, register_category, register_page

register_category(
    Category('personal',
             u'Personal preferences')
)

register_category(
    Category('settings',
             'Settings',
             after='personal')
)

from home_message import HomeMessage
register_page(HomeMessage)

from notes import Notes, NotesDisplay, NotesDisplayModeSwitch
register_page(Notes)
register_page(NotesDisplay)
register_page(NotesDisplayModeSwitch)

from restricted import HomeMessageRestricted, \
    HomeMessageRestrictedII, \
    HomeMessageRestrictedIII,\
    NotesRestricted
register_page(HomeMessageRestricted)
register_page(HomeMessageRestrictedII)
register_page(HomeMessageRestrictedIII)
register_page(NotesRestricted)

from defect import WrongCatIdHomeMessage, \
    ClashingWidgetIdHomeMessage,\
    NoWidgetIdPage
register_page(WrongCatIdHomeMessage)
register_page(ClashingWidgetIdHomeMessage)
register_page(NoWidgetIdPage)



# This module shows examples of what NOT to do when creating pages.

from Products.Five import BrowserView
from collective.mcp.samples.home_message import HomeMessage


class WrongCatIdHomeMessage(HomeMessage):
    """ This subclass uses an incorrect id when specifying
    the category (setings instead of settings)
    """
    category = 'setings'
    zcml_id = 'collective_mcp_home_message_defect1'
    widget_id = 'collective_mcp_home_message_defect1'
    title = 'Home message (in wrong category)'


class ClashingWidgetIdHomeMessage(HomeMessage):
    """ This subclass is using the same ``widget_id`` than
    HomeMessage.
    It will not be displayed in the control panel and a warning will
    be shown in the logs.
    """
    zcml_id = 'collective_mcp_home_message_defect2'
    title = 'Home message (defect due to widget_id clash)'


class NoWidgetIdPage(BrowserView):
    """ Do not try to create a page that do not subclass
    collective.mcp.browser.control_panel_page.ControlPanelPage, you might
    miss important stuff, like the widget_id for example.
    """
    category = 'settings'
    zcml_id = 'collective_mcp_home_message_defect3'
    title = 'Home message (without widget_id)'



from HTMLParser import HTMLParser


class McpHtmlParser(HTMLParser):
    """ The goal of this parser is to
    extract the usefull part of the browser contents:
     - the list of control panel categories available
     - the list of pages displayed
     - the list of buttons available for multi-objects pages
    """

    def __init__(self, *args, **kwargs):
        HTMLParser.__init__(self, *args, **kwargs)

        # True when a div with id 'subpage' is found in
        # the container.
        self.subpage_rendered = False

        # List of buttons rendered,
        # Each button is registered as a dictionnary:
        # {'href': ..., 'img': ...}
        self.multi_object_buttons = []

        # ist of objects in the list,
        # Each is registered as a dictionnary:
        # {'href':..., 'text': ..., 'current': True/False}
        self.multi_objects = []

        # List if items on the home page.
        # [{'title': 'Settings',
        #   'pages': [{'href': ..., 'title': ..., 'icon': ...},
        #             {'href': ..., 'title': ..., 'icon': ...},
        #             ...]},
        #  {...}]
        self.home_page = []

        # Classes applied on the #subpage div
        self.subpage_classes = ''

        # List of nodes encountered
        self.path = []

        # A few flags: they should all be reseted to False at the end.
        self.in_control_panel = False
        self.in_control_panel_header = False
        self.in_control_panel_content = False
        self.in_objects_list = False
        self.in_objects = False
        self.in_buttons = False
        self.in_menu = False

    def print_home_page(self):
        print 'Home page'
        for item in self.home_page:
            print ' - %s:' % item.get('title', 'No title')

            for p in item.get('pages', []):
                print '   + %s (%s, %s)' % (
                    p.get('title', 'no title'),
                    p.get('icon', 'no icon'),
                    p.get('href', 'no href'))

    def print_objects(self):
        print 'Objects'
        for o in self.multi_objects:
            print '%s %s (%s)' % (
                o.get('current', False) and '-->' or '   ',
                o.get('text', 'No text'),
                o.get('href', 'No href'))

    def print_buttons(self):
        print 'Buttons'
        for b in self.multi_object_buttons:
            print ' - %s (%s)' % (
                b.get('img', 'no image'),
                b.get('href', 'no href'))

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)

        tag_id = attrs.get('id', '')
        tag_classes = attrs.get('class', '')
        self.path.append({'tag': tag,
                          'id': tag_id,
                          'class': tag_classes})

        if tag == 'div':
            if tag_id == 'control_panel_container':
                self.in_control_panel = True
            elif tag_id == 'control_panel_header':
                self.in_control_panel_header = True
            elif tag_id == 'control_panel_content':
                self.in_control_panel_content = True
            elif tag_id == 'subpage':
                self.subpage_rendered = True
                self.subpage_classes = tag_classes
            elif tag_id == 'objects_list':
                self.in_objects_list = True

        if tag == 'ul':
            if 'objects' in tag_classes and self.in_objects_list:
                self.in_objects = True
            if 'buttons' in tag_classes and self.in_objects_list:
                self.in_buttons = True
            if tag_classes == 'menu' and self.in_control_panel_content:
                self.in_menu = True

        if tag == 'li':
            if self.in_objects:
                current = 'current' in tag_classes
                self.multi_objects.append({'current': current})

        if tag == 'a':
            if self.in_objects and self.multi_objects:
                self.multi_objects[-1]['href'] = attrs.get('href', '')
            elif self.in_buttons:
                self.multi_object_buttons.append(
                    {'href': attrs.get('href', '')})
            elif self.in_menu and self.home_page:
                self.home_page[-1]['pages'].append(
                    {'href': attrs.get('href', '')})

    def handle_endtag(self, tag):
        if not self.path:
            return

        t = self.path.pop()
        tag_id = t.get('id', '')
        tag_classes = t.get('class', '')

        if t['tag'] == 'div':
            if tag_id == 'control_panel_container':
                self.in_control_panel = False
            elif tag_id == 'control_panel_header':
                self.in_control_panel_header = False
            elif tag_id == 'control_panel_content':
                self.in_control_panel_content = False
            elif tag_id == 'subpage':
                self.subpage_rendered = False
            elif tag_id == 'objects_list':
                self.in_objects_list = False

        if t['tag'] == 'ul':
            if 'objects' in tag_classes and self.in_objects_list:
                self.in_objects = False
            if 'buttons' in tag_classes and self.in_objects_list:
                self.in_buttons = False
            if tag_classes == 'menu' and self.in_control_panel_content:
                self.in_menu = False

    def handle_startendtag(self, tag, attrs):
        attrs = dict(attrs)

        if tag == 'img':
            if self.in_buttons and self.multi_object_buttons:
                self.multi_object_buttons[-1]['img'] = attrs.get('src', '')
            elif self.in_menu and self.home_page and self.home_page[-1]['pages']:
                self.home_page[-1]['pages'][-1]['icon'] = attrs.get('src', '')

    def handle_data(self, data):
        if not self.path:
            return
        tag = self.path[-1]

        if tag['tag'] == 'a':
            if self.in_objects and self.multi_objects:
                self.multi_objects[-1]['text'] = data
            elif self.in_buttons and self.multi_object_buttons:
                self.multi_object_buttons[-1]['text'] = data

        if tag['tag'] == 'span':
            if self.in_menu:
                if 'spacer' in tag['class']:
                    self.home_page.append({'title': data, 'pages': []})
                elif self.home_page and self.home_page[-1]['pages']:
                    self.home_page[-1]['pages'][-1]['title'] = data



from collective.multimodeview.browser import MultiModeViewlet


class ControlPanelPage(MultiModeViewlet):
    # The category the page belongs to.
    category = ''

    # Relative URL of the icon shown in the main menu.
    icon = ''

    # Name declared in the ZCML
    zcml_id = ''

    # Tells if a page manages multiple objects.
    # In this case you need to define list_objects
    multi_objects = False

    # The mode used when clicking on an object
    # in the object list.
    multi_objects_default_mode = 'edit'

    # The default buttons displayed at the bottom
    # of the object list.
    # To use them, you need to have a 'add' and
    # 'delete' mode.
    multi_objects_buttons = ['add', 'delete']

    # A list of buttons that are appended after the default one.
    # Each button definition is a dictionnary with the following
    # keys:
    # - title: the string shown
    # - mode: the mode activated when clicking the button
    # - icon: the icon used the replace the title (not needed)
    multi_objects_extra_buttons = []

    # Before and after can be used to specify the position of the page in the
    # menu.
    # The value must correspond to the 'widget_id' attribute of the page that
    # is used for ordering.
    before = None
    after = None

    # You can define here lists of extra CSS/Javascript files for this page.
    extra_js = []
    extra_css = []

    # Defines a list of CSS class that will be added to the subpage container.
    sub_page_classes = []

    @property
    def cancel_mode(self):
        if 'back' in self.modes:
            return 'back'
        return self.default_mode

    def __call__(self):
        self.on_call()
        return self.index()

    def list_objects(self):
        """ Returns a list of dictionnaries.
        Dictionnaries must have two keys:
         - id: which is the object id (at least
           something you can use to fetch id, it
           can be UID or anything else)
         - title: which is what is displayed
           in the left column.
        """
        return []

    def current_object_id(self):
        """ Returns the id of the object currently
        manager or None.
        It first checks for a 'obj_id' paramter
        in the request.
        If not (which is the case when you open the page
        for the first time), it returns the id object
        of the first object.
        """
        form = self.request.form
        if form.get('obj_id', None):
            return form.get('obj_id')

        objects = self.list_objects()
        if objects and self.mode != 'add':
            return objects[0]['id']

        return None

    def make_link(self, mode, extra_params=None):
        """ Removes the unneeded '#widget_id' at the end of the link.
        """
        link = super(ControlPanelPage, self).make_link(mode, extra_params)
        return link.replace('#%s' % self.widget_id, '')

    def page_classes(self):
        """ Generates the CSS classes subpage container..
        """
        classes = self.sub_page_classes[:]

        widget_id = self.request.get('widget_id', None)
        if widget_id:
            classes.append('mcp_widget_%s' % widget_id)

        if self.multi_objects:
            classes.append('right_column')

        return ' '.join(classes)



from collective.mcp.browser.control_panel_page import ControlPanelPage


class HomeMessage(ControlPanelPage):
    category = 'settings'
    zcml_id = 'collective_mcp_home_message'
    widget_id = 'collective_mcp_home_message'
    icon = "++resource++collective_mcp_home.png"
    title = 'Home message'

    modes = {'back': {},
             'default': {'submit_label': 'Update home message',
                         'success_msg': 'The home message has been updated'}}
    default_mode = 'default'

    @property
    def notes_view(self):
        return self.context.restrictedTraverse(
            '@@mcp_multimodeview_notes_sample')

    def _check_default_form(self):
        return True

    def _process_default_form(self):
        self.notes_view.set_home_message(
            self.request.form.get('msg', ''))
        return 'back'



from collective.multimodeview.samples.notes_view import NotesView
from persistent.dict import PersistentDict


class McpNotesView(NotesView):
    """ Provides some extra actions than the one defined
    in multimodeview.
    """
    display_zones_list = ['home', 'news', 'control panel']

    def get_display_zones(self):
        metadata = self._get_metadata()

        display_zones = metadata.get('display_zones',
                                     None)
        if display_zones is None:
            metadata['display_zones'] = PersistentDict()
            display_zones = metadata['display_zones']

        return display_zones

    def set_display_zones(self, note_id, zones):
        dzones = self.get_display_zones()
        dzones[note_id] = zones



class Category(object):
    def __init__(self, cat_id, title, before=None, after=None):
        self.id = cat_id
        self.title = title

        self.before = before
        self.after = after

    def __repr__(self):
        return 'Category: %s' % self.id

    def getTitle(self):
        return self.title



from setuptools import setup, find_packages
import os


def get_file_contents(*filenames):
    file_path = os.path.join(*filenames)
    this_file = open(file_path)
    contents = this_file.read().strip()
    this_file.close()
    return contents

history = get_file_contents('CHANGES.rst')
readme = get_file_contents('README.rst')
main = get_file_contents('collective', 'mcp', 'doc', 'main.rst')
long_description = "%s\n\n\n%s\n\n\n%s" % (readme, main, history)

setup(name='collective.mcp',
      version='0.5',
      description="Macish Control Panel for Plone.",
      long_description=long_description,
      # Get more strings from
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
          "Framework :: Plone",
          "Framework :: Plone :: 3.3",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.4",
          ],
      keywords='mac control panel',
      author='Zest Software',
      author_email='info@zestsoftware.nl',
      url='https://github.com/zestsoftware/collective.mcp',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['collective'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          # -*- Extra requirements: -*-
          'collective.multimodeview'
      ],
      entry_points="""
      # -*- Entry points: -*-

      [z3c.autoinclude.plugin]
      target = plone
      """,
      )



import logging
logger = logging.getLogger('collective.mcp')


class SortableObj(object):
    """ That class is only here for testing purpose.
    """

    def __init__(self, obj_id, before=None, after=None):
        self.id = obj_id
        self.before = before
        self.after = after

    def __repr__(self):
        return '<Sortable object: %s>' % self.id


class SortedList(list):
    """ This type of list adds a 'add' method, that will insert
    the object at the correct position, based on 'before'/'after'
    attributes of the object.
    """

    def __init__(self, *args, **kwargs):
        if 'id_attr' in kwargs:
            id_attribute = kwargs['id_attr']
            del kwargs['id_attr']
        else:
            id_attribute = 'id'
        super(SortedList, self).__init__(*args, **kwargs)
        self.id_attribute = id_attribute

    def add(self, obj):
        """ This method adds an object to the list and  will place it according
        to the 'before'/'after' attribute.

        If the object do not specify a position, it is
        appended to the list.
        >>> l = SortedList()
        >>> l.add(SortableObj('bla'))
        >>> l
        [<Sortable object: bla>]

        >>> l.add(SortableObj('bli'))
        >>> l
        [<Sortable object: bla>, <Sortable object: bli>]

        If an object just specify the 'after' property,
        it will be placed just after the object asked.
        >>> l.add(SortableObj('blu',
        ...                   after='bla'))
        >>> l
        [<Sortable object: bla>, <Sortable object: blu>, <Sortable object: bli>]

        Same thing for the 'before' property
        >>> l.add(SortableObj('blo',
        ...                   before='blu'))
        >>> l
        [<Sortable object: bla>, <Sortable object: blo>, <Sortable object: blu>,
         <Sortable object: bli>]

        And of course, it it specifies both, we
        will try to place it in between.
        >>> l.add(SortableObj('ble',
        ...                   before='bli',
        ...                   after='blu'))
        >>> l
        [<Sortable object: bla>, <Sortable object: blo>, <Sortable object: blu>,
         <Sortable object: ble>, <Sortable object: bli>]

        If existing objects are specifying unknown relation, they
        will be taken into account.
        >>> l.add(SortableObj('after_a',
        ...                   after='a'))
        >>> l
        [<Sortable object: bla>, <Sortable object: blo>, <Sortable object: blu>,
         <Sortable object: ble>, <Sortable object: bli>, <Sortable object: after_a>]
        >>> l.add(SortableObj('a'))
        >>> l
        [<Sortable object: bla>, <Sortable object: blo>, <Sortable object: blu>,
         <Sortable object: ble>, <Sortable object: bli>, <Sortable object: a>,
         <Sortable object: after_a>]

        If needed, the objects in the list can also be re-positionned.
        >>> l1 = SortedList()
        >>> l1.add(SortableObj('c', after='b'))
        >>> l1.add(SortableObj('a', before='b'))
        >>> l1
        [<Sortable object: c>, <Sortable object: a>]
        >>> l1.add(SortableObj('b'))
        >>> l1
        [<Sortable object: a>, <Sortable object: b>, <Sortable object: c>]
        """
        try:
            obj_id = getattr(obj, self.id_attribute)
        except AttributeError:
            return

        after_position = None
        before_position = None

        for index, o in enumerate(self):
            o_id = getattr(o, self.id_attribute)

            # We always increment after_position
            # as long as we find a reason to continue
            if o.before == obj_id or \
               obj.after == o_id:
                after_position = index + 1

            # Once a before_position has been found, it is
            # never changed (which is logical)
            if before_position is None and \
               (o.after == obj_id or
                    obj.before == o_id):
                before_position = index

        if after_position is None and \
           before_position is None:
            # Well, we don't really have a clue about what to do, let's
            # put it at the end.
            self.append(obj)

        elif after_position is None and \
                before_position is not None:
            # We place it just before the asked position.
            self.insert(before_position, obj)

        elif before_position is None and \
                after_position is not None:
            # Ok, we place it just after the position we found.
            self.insert(after_position, obj)

        elif after_position <= before_position:
            # Logically, if we arrived here, both objects
            # are not None.
            # This case is the "classic" case, the object has to
            # be placed between two elements.
            self.insert(after_position, obj)

        else:
            # Ok, that case sucks, it means we need to reorganize
            # the list.

            # First, we check that there is no possible clash (typically,
            # the element must be before AND after another one)
            all_before = self.find_all_before([obj])
            all_after = self.find_all_after([obj])
            conflicts = [x for x in self
                         if x in all_before and x in all_after]
            if conflicts:
                msg = 'Could not place object %s, as those objects ' +\
                      'must be before ADN after it: %s'
                logger.warn(msg % (obj, conflicts))
                self.append(obj)
                return

            # Now what we will do is place objects in the 'all_before' list
            # just before the first element of 'all_after'.
            position = self.find_first_element(all_after)
            self.insert(position, obj)
            for o in sorted(all_before,
                            reverse=True,
                            key=lambda x: self.index(x)):
                self.remove(o)
                self.insert(position, o)

    def _find_all(self, objects, match):
        """
        """
        path = objects
        found = []

        while path:
            current = path.pop(0)
            for o in self:
                if o == current:
                    # We'll try to avoid recursion.
                    continue

                if match(current, o) and o not in found:
                    found.append(o)
                    path.append(o)

        return found

    def find_all_before(self, objects):
        """ Finds every objects that are supposed to be
        placed before any objects in the list.

        It finds direct relations (here 'a' and 'b'
        explicitely say that they are before 'c')
        >>> l = SortedList()
        >>> l.add(SortableObj('a', before = 'c'))
        >>> l.add(SortableObj('b', before = 'c'))
        >>> l.add(SortableObj('c', before = 'd'))
        >>> l
        [<Sortable object: a>, <Sortable object: b>, <Sortable object: c>]
        >>> l.find_all_before([l[2]])
        [<Sortable object: a>, <Sortable object: b>]

        And also indirect ones ('a' and 'b' appears
        as they are before 'c' who is before 'd')
        >>> l.add(SortableObj('d'))
        >>> l.find_all_before([l[-1]])
        [<Sortable object: c>, <Sortable object: a>, <Sortable object: b>]

        And if nothing can be found, well if returns an empty list.
        >>> l.find_all_before([l[0]])
        []
        """
        def compare(current, o):
            current_id = getattr(current, self.id_attribute)
            o_id = getattr(o, self.id_attribute)

            return o.before == current_id or \
                current.after == o_id

        return self._find_all(objects, compare)

    def find_all_after(self, objects):
        """ Same principle as 'find_add_before', except
        if seeks for all objects that are supposed to be placed
        after any object in the objects list.

        It finds direct relations (here 'b' and 'c'
        explicitely say that they are before 'a')
        >>> l = SortedList()
        >>> l.add(SortableObj('c', after = 'a'))
        >>> l.add(SortableObj('b', after = 'a'))
        >>> l.add(SortableObj('a', after = '0'))
        >>> l
        [<Sortable object: a>, <Sortable object: c>, <Sortable object: b>]
        >>> l.find_all_after([l[0]])
        [<Sortable object: c>, <Sortable object: b>]

        And also indirect ones ('b' and 'c' appears
        as they are after 'a' who is after '0')
        >>> l.add(SortableObj('0'))
        >>> l.find_all_after([l[0]])
        [<Sortable object: a>, <Sortable object: c>, <Sortable object: b>]

        >>> l.find_all_after([l[-1]])
        []
        """
        def compare(current, o):
            current_id = getattr(current, self.id_attribute)
            o_id = getattr(o, self.id_attribute)

            return o.after == current_id or \
                current.before == o_id

        return self._find_all(objects, compare)

    def find_first_element(self, objects):
        """ Finds the position of the first
        element listed in objects.

        >>> l = SortedList()
        >>> l.add(SortableObj('a'))
        >>> l.add(SortableObj('b'))
        >>> l.add(SortableObj('c'))
        >>> l.add(SortableObj('d'))
        >>> l.add(SortableObj('e'))
        >>> l
        [<Sortable object: a>,
         <Sortable object: b>,
         <Sortable object: c>,
         <Sortable object: d>,
         <Sortable object: e>]

        >>> l.find_first_element([l[3], l[2], l[4]])
        2

        >>> l.find_first_element(['a', 'b', 'c'])
        """
        for index, obj in enumerate(self):
            if obj in objects:
                return index



from collective.mcp.category import Category
from collective.mcp.sorted_list import SortedList
from zope.i18nmessageid import MessageFactory
import logging

logger = logging.getLogger('collective.mcp')
McpMessageFactory = MessageFactory(u'collective.mcp')

categories = SortedList(id_attr='id')
pages = SortedList(id_attr='widget_id')


def register_category(cat):
    if isinstance(cat, Category):
        if cat.id in [c.id for c in categories]:
            logger.info('Category id "%s" already exists' % cat.id)
        else:
            categories.add(cat)
    else:
        raise TypeError('Expected Category object, found %s' % type(cat))


def register_page(p):
    if p.category not in categories:
        logger.info('There is no "%s" category registered' % p.category)
    pages.add(p)


def initialize(context):
    """Initializer called when used as a Zope 2 product."""



from Testing import ZopeTestCase as ztc
import doctest
import unittest

OPTIONFLAGS = (doctest.ELLIPSIS |
               doctest.NORMALIZE_WHITESPACE)


def test_suite():
    return unittest.TestSuite([
        ztc.ZopeDocFileSuite(
            'tests/coverage.txt',
            package='collective.mcp',
            optionflags=OPTIONFLAGS),
        ztc.ZopeDocFileSuite(
            'tests/sorted_list.txt',
            package='collective.mcp',
            optionflags=OPTIONFLAGS),
        doctest.DocTestSuite(
            module='collective.mcp.sorted_list',
            optionflags=OPTIONFLAGS),
    ])



from collective.mcp.browser.control_panel_page import ControlPanelPage


class Notes(ControlPanelPage):
    category = 'settings'
    zcml_id = 'collective_mcp_notes'
    widget_id = 'collective_mcp_notes'
    icon = "++resource++collective_mcp_notes.png"
    title = 'Notes'

    modes = {'add': {'success_msg': 'The note has been added',
                     'error_msg': 'Impossible to add a note: please correct the form',
                     'submit_label': 'Add note'},
             'edit': {'success_msg': 'The note has been edited',
                      'submit_label': 'Edit note'},
             'delete': {'success_msg': 'The note has been deleted',
                        'submit_label': 'Delete note'}
             }
    default_mode = 'edit'
    multi_objects = True

    @property
    def notes_view(self):
        return self.context.restrictedTraverse(
            '@@mcp_multimodeview_notes_sample')

    def list_objects(self):
        notes = self.notes_view.get_notes()

        return [{'id': note_id, 'title': note_text}
                for note_id, note_text in enumerate(notes)
                if note_text]

    def _get_note_id(self):
        notes = self.notes_view.get_notes()
        note_id = self.current_object_id()

        try:
            note_id = int(note_id)
        except:
            # This should not happen, something wrong happened
            # with the form.
            return

        if note_id < 0 or note_id >= len(notes):
            # Again, something wrong hapenned.
            return

        if notes[note_id] is None:
            # This note has been deleted, nothing should be done
            # with it.
            return

        return note_id

    def get_note_title(self):
        """ Returns the title of the note currently edited.
        """
        if self.errors:
            return self.request.form.get('title')

        if self.is_add_mode:
            return ''

        note_id = self._get_note_id()
        if note_id is None:
            # This should not happen.
            return ''

        return self.notes_view.get_notes()[note_id]

    def _check_add_form(self):
        if not self.request.form.get('title'):
            self.errors['title'] = 'You must provide a title'

        return True

    def _check_edit_form(self):
        if self._get_note_id() is None:
            return

        return self._check_add_form()

    def _check_delete_form(self):
        return self._get_note_id() is not None

    def _process_add_form(self):
        self.notes_view.add_note(self.request.form.get('title'))
        self.request.form['obj_id'] = len(self.notes_view.get_notes()) - 1

    def _process_edit_form(self):
        self.notes_view.edit_note(
            self._get_note_id(),
            self.request.form.get('title'))

    def _process_delete_form(self):
        self.notes_view.delete_note(self._get_note_id())
        self.request.form['obj_id'] = None


class NotesDisplay(Notes):
    """ More or less the same than the previous one, except it
    has an extra 'display' mode.
    """
    zcml_id = 'collective_mcp_notes_display'
    widget_id = 'collective_mcp_notes_display'
    title = 'Displayed notes'

    modes = {'add': {'success_msg': 'The note has been added',
                     'error_msg': 'Impossible to add a note: please correct the form',
                     'submit_label': 'Add note'},
             'edit': {'success_msg': 'The note has been edited',
                      'submit_label': 'Edit note'},
             'display': {'success_msg': 'Display zones for the note have been updated',
                         'submit_label': 'Set display zones'},
             'delete': {'success_msg': 'The note has been deleted',
                        'submit_label': 'Delete note'}
             }

    multi_objects_extra_buttons = [
        {'mode': 'display',
         'title': 'Change display zones',
         'icon': '++resource++display_button.png'}]

    def get_zones(self):
        zones = []

        note_id = self._get_note_id()
        display_zones = self.notes_view.get_display_zones()
        note_display_zones = display_zones.get(note_id, [])

        for zone in self.notes_view.display_zones_list:
            zones.append({'value': zone,
                          'checked': zone in note_display_zones})
        return zones

    def _check_display_form(self):
        return True

    def _process_display_form(self):
        zones = self.request.form.get('zones', [])

        if isinstance(zones, str):
            # Only one zone has been checked.
            zones = [zones]

        self.notes_view.set_display_zones(
            self._get_note_id(),
            zones)


class NotesDisplayModeSwitch(NotesDisplay):
    """ Same principle than the previous one.
    It adds some extra parameters when defining the mode to declare
    to which mode switch after adding a note or after cancelling.
    """
    zcml_id = 'collective_mcp_notes_mode_switch'
    widget_id = 'collective_mcp_notes_mode_switch'
    title = 'Mode switching notes'

    modes = {'add': {'success_msg': 'The note has been added',
                     'error_msg': 'Impossible to add a note: please correct the form',
                     'submit_label': 'Add note',
                     'cancel_mode': 'add',
                     'success_mode': 'display'},
             'edit': {'success_msg': 'The note has been edited',
                      'submit_label': 'Edit note'},
             'display': {'success_msg': 'Display zones for the note have been updated',
                         'submit_label': 'Set display zones'},
             'delete': {'success_msg': 'The note has been deleted',
                        'submit_label': 'Delete note'}
             }
